from .dimred import reduce, reduce_scanpy
from .generate import pancakes
from .integrate import correct, correct_scanpy
from .plotting import ordered_matrixplot
from .score import get_binom_scores, info_score
from .utils import metagene_loadings
